'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.default = function (babel) {
  var t = babel.types;


  function isProgram(path) {
    return path.isProgram();
  }

  /**
   * 判断 function 函数和箭头函数
   */
  function isAnyFunctionExpression() {
    return t.isFunctionExpression.apply(t, arguments) || t.isArrowFunctionExpression.apply(t, arguments);
  }

  /**
   * 判断节点是否是 action
   */
  function isAction(node, actionIdentifier, mobxNamespaceIdentifier) {
    return actionIdentifier && t.isIdentifier(node, { name: actionIdentifier }) || t.isIdentifier(node, { name: 'runInAction' }) || t.isMemberExpression(node) && t.isIdentifier(node.object, { name: 'action' }) && t.isIdentifier(node.property, { name: "bound" }) || mobxNamespaceIdentifier && t.isMemberExpression(node) && t.isIdentifier(node.object, { name: mobxNamespaceIdentifier }) && t.isIdentifier(node.property, { name: "action" });
  }

  /**
   * 遍历被 action 包裹的函数及箭头函数节点
   */
  var traverseActionBody = _defineProperty({}, "FunctionExpression|ArrowFunctionExpression", function FunctionExpressionArrowFunctionExpression(path) {
    var actionIdentifier = this.actionIdentifier;
    var mobxNamespaceIdentifier = this.mobxNamespaceIdentifier;
    path.traverse(traverseActionBody, { actionIdentifier: actionIdentifier, mobxNamespaceIdentifier: mobxNamespaceIdentifier });
    path.skip();
    // if current node parent is call expression and this call is action call, skip wrapping
    if (t.isCallExpression(path.parentPath.node) && isAction(path.parentPath.node.callee, actionIdentifier, mobxNamespaceIdentifier)) {
      return;
    }
    path.replaceWith(t.CallExpression(this.actionIdentifier ? t.Identifier(this.actionIdentifier) : t.MemberExpression(t.Identifier(this.mobxNamespaceIdentifier), t.Identifier("action")), [path.node]));
  });

  /**
   * 遍历兄弟节点
   */
  var traverseSibling = _defineProperty({
    CallExpression: function CallExpression(path) {
      // 查询函数内部的方法节点
      var node = path.node;
      var actionIdentifier = this.actionIdentifier;
      var mobxNamespaceIdentifier = this.mobxNamespaceIdentifier;
      if (isAction(node.callee, actionIdentifier, mobxNamespaceIdentifier)) {
        // 调用isAction 判断是否是 action 节点
        if (node.arguments.length === 1) {
          path.get('arguments.0').traverse(traverseActionBody, { actionIdentifier: actionIdentifier, mobxNamespaceIdentifier: mobxNamespaceIdentifier });
          path.skip();
        } else if (node.arguments.length === 2) {
          path.get('arguments.1').traverse(traverseActionBody, { actionIdentifier: actionIdentifier, mobxNamespaceIdentifier: mobxNamespaceIdentifier });
          path.skip();
        }
      }
    }
  }, "ClassMethod|ClassProperty", function ClassMethodClassProperty(path) {
    var actionIdentifier = this.actionIdentifier;
    var mobxNamespaceIdentifier = this.mobxNamespaceIdentifier;
    var explicitClasses = this.classes;
    var classDeclaration = path.findParent(function (p) {
      return p.isClassDeclaration();
    });

    // 判断 constructor 是否使用了 makeAutoObservable
    if (path.node.key.name === "constructor") {
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = path.container[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var node = _step.value;

          if (node.kind === "constructor") {
            var _iteratorNormalCompletion2 = true;
            var _didIteratorError2 = false;
            var _iteratorError2 = undefined;

            try {
              for (var _iterator2 = node.body.body[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                var item = _step2.value;

                this.hasUseMakeAutoObservable = t.isIdentifier(item.expression.callee, { name: 'makeAutoObservable' });
              }
            } catch (err) {
              _didIteratorError2 = true;
              _iteratorError2 = err;
            } finally {
              try {
                if (!_iteratorNormalCompletion2 && _iterator2.return) {
                  _iterator2.return();
                }
              } finally {
                if (_didIteratorError2) {
                  throw _iteratorError2;
                }
              }
            }
          }
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    }
    // 如果使用了 makeAutoObservable 并且是类的方法
    if (this.hasUseMakeAutoObservable && t.isClassMethod(path.node, { kind: 'method' })) {
      if (t.isClassMethod(path.node)) {
        path.get('body').traverse(traverseActionBody, { actionIdentifier: actionIdentifier, mobxNamespaceIdentifier: mobxNamespaceIdentifier });
        path.skip();
      }
    } else if (this.hasUseMakeAutoObservable && (t.isFunctionExpression(path.node.value) || t.isArrowFunctionExpression(path.node.value))) {
      path.get('value').traverse(traverseActionBody, { actionIdentifier: actionIdentifier, mobxNamespaceIdentifier: mobxNamespaceIdentifier });
    }

    if (explicitClasses && t.isIdentifier(classDeclaration.node.id) && void 0 !== explicitClasses[classDeclaration.node.id.name] && t.isClassMethod(path.node) && t.isIdentifier(path.node.key) && (
    // all code inside constructor should be handled as action too, because it could contain other action creations
    path.node.key.name === "constructor" || void 0 !== explicitClasses[classDeclaration.node.id.name][path.node.key.name])) {
      if (path.node.key.name === "constructor") {
        path.get('body').get('body').forEach(function (cPath) {
          if (cPath.isExpressionStatement()) {
            var exprPath = cPath.get('expression');
            if (exprPath.isAssignmentExpression() && exprPath.get('operator').node === '=') {
              var leftPath = exprPath.get('left');
              var rightPath = exprPath.get('right');
              if (leftPath.isMemberExpression() && leftPath.get('object').isThisExpression() && leftPath.get('property').isIdentifier() && leftPath.get('property').node.name in explicitClasses[classDeclaration.node.id.name] && (rightPath.isArrowFunctionExpression() || rightPath.isFunctionExpression())) {
                rightPath.get('body').traverse(traverseActionBody, { actionIdentifier: actionIdentifier, mobxNamespaceIdentifier: mobxNamespaceIdentifier });
              }
            }
          }
        });
      } else {
        path.get('body').traverse(traverseActionBody, { actionIdentifier: actionIdentifier, mobxNamespaceIdentifier: mobxNamespaceIdentifier });
      }
      path.skip();
    } else if (path.node.decorators) {
      var _iteratorNormalCompletion3 = true;
      var _didIteratorError3 = false;
      var _iteratorError3 = undefined;

      try {
        for (var _iterator3 = path.node.decorators[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
          var _ref2 = _step3.value;
          var expression = _ref2.expression;

          if (isAction(expression, actionIdentifier, mobxNamespaceIdentifier) || t.isCallExpression(expression) && isAction(expression.callee, actionIdentifier, mobxNamespaceIdentifier)) {
            if (t.isClassMethod(path.node)) {
              path.get('body').traverse(traverseActionBody, { actionIdentifier: actionIdentifier, mobxNamespaceIdentifier: mobxNamespaceIdentifier });
              path.skip();
            } else if (t.isClassProperty(path.node)) {
              path.get('value').traverse(traverseActionBody, { actionIdentifier: actionIdentifier, mobxNamespaceIdentifier: mobxNamespaceIdentifier });
              path.skip();
            }
          }
        }
      } catch (err) {
        _didIteratorError3 = true;
        _iteratorError3 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion3 && _iterator3.return) {
            _iterator3.return();
          }
        } finally {
          if (_didIteratorError3) {
            throw _iteratorError3;
          }
        }
      }
    }
  });

  return {
    name: "mobx-deep-action-transform",
    visitor: {
      Program: function Program(path, state) {
        var actionIdentifier = void 0;
        var mobxNamespaceIdentifier = void 0;
        var hasUseMakeAutoObservable = false; // 判断是否使用了 MakeAutoObservable
        var tslibNamespaceIdentifier = void 0;
        var mobxPackage = state.opts && state.opts["mobx-package"] || "mobx";
        path.traverse({
          ImportDeclaration: function ImportDeclaration(path) {
            if (path.node.source.value === mobxPackage) {
              // 判断是否导入了 mobx 包
              var hasImportAction = false;
              var hasImportMakeAutoObservable = false;
              // 循环遍历，确认导入了什么模块
              var _iteratorNormalCompletion4 = true;
              var _didIteratorError4 = false;
              var _iteratorError4 = undefined;

              try {
                for (var _iterator4 = path.node.specifiers[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
                  var specifier = _step4.value;

                  if (t.isImportNamespaceSpecifier(specifier)) {
                    mobxNamespaceIdentifier = specifier.local.name; // 确认是导入了 mobx 包还是 mobx-package
                  } else if (specifier.imported.name === "action") {
                    actionIdentifier = specifier.local.name;
                    hasImportAction = true;
                  } else if (specifier.imported.name === "makeAutoObservable") {
                    hasImportMakeAutoObservable = true;
                  }
                }

                // 导入了 makeAutoObservable 但是没有导入 action，自动把 action 加进去
              } catch (err) {
                _didIteratorError4 = true;
                _iteratorError4 = err;
              } finally {
                try {
                  if (!_iteratorNormalCompletion4 && _iterator4.return) {
                    _iterator4.return();
                  }
                } finally {
                  if (_didIteratorError4) {
                    throw _iteratorError4;
                  }
                }
              }

              if (hasImportMakeAutoObservable && !hasImportAction) {
                var specifiers = [t.importSpecifier(t.identifier('action'), t.identifier('action'))];
                var importDeclaration = t.importDeclaration(specifiers, t.stringLiteral("mobx"));
                var program = path.findParent(isProgram);
                program.unshiftContainer("body", importDeclaration);
              }
            }
            if (path.node.source.value === "tslib") {
              var _iteratorNormalCompletion5 = true;
              var _didIteratorError5 = false;
              var _iteratorError5 = undefined;

              try {
                for (var _iterator5 = path.node.specifiers[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
                  var _specifier = _step5.value;

                  if (t.isImportNamespaceSpecifier(_specifier)) {
                    tslibNamespaceIdentifier = _specifier.local.name;
                  }
                }
              } catch (err) {
                _didIteratorError5 = true;
                _iteratorError5 = err;
              } finally {
                try {
                  if (!_iteratorNormalCompletion5 && _iterator5.return) {
                    _iterator5.return();
                  }
                } finally {
                  if (_didIteratorError5) {
                    throw _iteratorError5;
                  }
                }
              }
            }
          }
        });
        var context = { actionIdentifier: actionIdentifier, mobxNamespaceIdentifier: mobxNamespaceIdentifier, hasUseMakeAutoObservable: hasUseMakeAutoObservable };
        path.traverse(traverseSibling, context);
        var toTraverse = [];
        /**
         * Lookup for typescript decorators, and handle them separately
         */
        path.traverse({
          CallExpression: function CallExpression(path) {
            // 查询函数内部的方法节点
            var node = path.node;
            if (t.isMemberExpression(node.callee) && t.isIdentifier(node.callee.object, { name: tslibNamespaceIdentifier }) && t.isIdentifier(node.callee.property, { name: "__decorate" }) && node.arguments.length === 4 && t.isArrayExpression(node.arguments[0]) && node.arguments[0].elements.some(function (e) {
              return t.isIdentifier(e, { name: actionIdentifier }) || t.isMemberExpression(e) && t.isIdentifier(e.object, { name: mobxNamespaceIdentifier }) && t.isIdentifier(e.property, { name: "action" }) || t.isCallExpression(e) && t.isIdentifier(e.callee, { name: actionIdentifier }) || t.isCallExpression(e) && t.isMemberExpression(e.callee) && t.isIdentifier(e.callee.object, { name: mobxNamespaceIdentifier }) && t.isIdentifier(e.callee.property, { name: "action" });
            }) && t.isMemberExpression(node.arguments[1]) && t.isIdentifier(node.arguments[1].property, { name: "prototype" }) && t.isStringLiteral(node.arguments[2])) {
              var className = node.arguments[1].object.name;
              var methodName = node.arguments[2].value;
              var traversePath = path.getStatementParent().parentPath;
              var existsTraverseRequest = toTraverse.find(function (e) {
                return e.path === traversePath;
              });
              if (!existsTraverseRequest) {
                toTraverse.push({
                  path: traversePath,
                  classes: _defineProperty({}, className, _defineProperty({}, methodName, methodName))
                });
              } else {
                var existsClassRequest = existsTraverseRequest.classes[className];
                if (!existsClassRequest) {
                  existsTraverseRequest.classes[className] = _defineProperty({}, methodName, methodName);
                } else {
                  existsTraverseRequest.classes[className][methodName] = methodName;
                }
              }
            }
          }
        });
        toTraverse.forEach(function (_ref3) {
          var path = _ref3.path,
              classes = _ref3.classes;
          return path.traverse(traverseSibling, _extends({}, context, { classes: classes }));
        });
      }
    }
  };
};

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }